/*
** Copyright 1999-2001 by LivingLogic AG, Bayreuth, Germany.
** Copyright 1999-2001 by Walter Drwald
**
** All Rights Reserved
**
** Permission to use, copy, modify, and distribute this software and its documentation
** for any purpose and without fee is hereby granted, provided that the above copyright
** notice appears in all copies and that both that copyright notice and this permission
** notice appear in supporting documentation, and that the name of LivingLogic AG or
** the author not be used in advertising or publicity pertaining to distribution of the
** software without specific, written prior permission.
**
** LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
** INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
** LIVING LOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
** DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
** IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
** IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "Python.h"

static Py_UNICODE lt[] = { ((Py_UNICODE)'&'), ((Py_UNICODE)'l'), ((Py_UNICODE)'t'), ((Py_UNICODE)';') };
static Py_UNICODE gt[] = { ((Py_UNICODE)'&'), ((Py_UNICODE)'g'), ((Py_UNICODE)'t'), ((Py_UNICODE)';') };
static Py_UNICODE amp[] = { ((Py_UNICODE)'&'), ((Py_UNICODE)'a'), ((Py_UNICODE)'m'), ((Py_UNICODE)'p'), ((Py_UNICODE)';') };
static Py_UNICODE quot[] = { ((Py_UNICODE)'&'), ((Py_UNICODE)'q'), ((Py_UNICODE)'u'), ((Py_UNICODE)'o'), ((Py_UNICODE)'t'), ((Py_UNICODE)';') };

static Py_UNICODE lt2[] = { ((Py_UNICODE)'l'), ((Py_UNICODE)'t') };
static Py_UNICODE gt2[] = { ((Py_UNICODE)'g'), ((Py_UNICODE)'t') };
static Py_UNICODE amp2[] = { ((Py_UNICODE)'a'), ((Py_UNICODE)'m'), ((Py_UNICODE)'p') };
static Py_UNICODE quot2[] = { ((Py_UNICODE)'q'), ((Py_UNICODE)'u'), ((Py_UNICODE)'o'), ((Py_UNICODE)'t') };

#define COUNTOF(x) (sizeof(x)/sizeof((x)[0]))

struct Replacement
{
	Py_UNICODE  character;
	int         size;
	Py_UNICODE *replacement;
};

struct Replacement replacements[] =
{
	{'"', COUNTOF(quot2), quot2},
	{'&', COUNTOF(amp2), amp2},
	{'<', COUNTOF(lt2), lt2},
	{'>', COUNTOF(gt2), gt2},
	{'\0', 0, NULL}
};

#if PY_VERSION_HEX <= 0x020000F1
PyObject *PyObject_Unicode(PyObject *v)
{
	PyObject *res;

	if (v == NULL)
		res = PyString_FromString("<NULL>");
	else if (PyUnicode_Check(v))
	{
		Py_INCREF(v);
		return v;
	}
	else if (PyString_Check(v))
	{
		res = v;
		Py_INCREF(v);
	}
	else if (v->ob_type->tp_str != NULL)
		res = (*v->ob_type->tp_str)(v);
	else
	{
		PyObject *func;
		static PyObject *strstr;
		if (strstr == NULL)
		{
			strstr= PyString_InternFromString("__str__");
			if (strstr == NULL)
				return NULL;
		}
		if (!PyInstance_Check(v) || (func = PyObject_GetAttr(v, strstr)) == NULL)
		{
			PyErr_Clear();
			res = PyObject_Repr(v);
		}
		else
		{
			res = PyEval_CallObject(func, (PyObject *)NULL);
			Py_DECREF(func);
		}
	}
	if (res == NULL)
		return NULL;
	if (!PyUnicode_Check(res))
	{
		PyObject* str;
		str = PyUnicode_FromObject(res);
		Py_DECREF(res);
		if (str)
			res = str;
		else
			return NULL;
	}
	return res;
}
#endif

static PyUnicodeObject *escapeUnencodable(PyUnicodeObject *str, const char *encoding)
{
	PyObject *test = PyUnicode_AsEncodedString((PyObject *)str, encoding, NULL);

	if (test == NULL) /* an exception has occurred */
	{
		if (PyErr_ExceptionMatches(PyExc_UnicodeError)) /* OK, we try it character by character */
		{
			int i;
			int oldsize = PyUnicode_GET_SIZE(str);
			int newsize = 0;
			PyObject *result;
			Py_UNICODE *p;

			PyErr_Clear();

			/* determine the space, that we will need */
			for (i = 0; i < oldsize; ++i)
			{
				Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
				test = PyUnicode_Encode(&ch, 1, encoding, NULL);

				if (test == NULL) /* exception */
				{
					if (PyErr_ExceptionMatches(PyExc_UnicodeError)) /* we must create a character reference for it */
					{
						PyErr_Clear();

						if (ch < 10)
							newsize += 2+1+1;
						else if (ch < 100)
							newsize += 2+2+1;
						else if (ch < 1000)
							newsize += 2+3+1;
						else if (ch < 10000)
							newsize += 2+4+1;
						else
							newsize += 2+5+1;
					}
					else
						return NULL;
				}
				else /* we were able to encode this character, so use it */
				{
					++newsize;
					Py_DECREF(test);
				}
			}

			/* allocate what we calculated */
			result = PyUnicode_FromUnicode(NULL, newsize);

			if (result == NULL)
				return NULL;

			/* create the string */
			p = PyUnicode_AS_UNICODE(result);
			for (i = 0; i < oldsize; ++i)
			{
				Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
				test = PyUnicode_Encode(&ch, 1, encoding, NULL);

				if (test == NULL) /* exception */
				{
					if (PyErr_ExceptionMatches(PyExc_UnicodeError)) /* we must create a character reference for it */
					{
						PyErr_Clear();

						*p++ = (Py_UNICODE)'&';
						*p++ = (Py_UNICODE)'#';
						if (ch>10000)
						{
							*p++ = ((Py_UNICODE)'0') + ch/10000;
							ch = ch % 10000;
						}
						if (ch>1000)
						{
							*p++ = ((Py_UNICODE)'0') + ch/1000;
							ch = ch % 1000;
						}
						if (ch>100)
						{
							*p++ = ((Py_UNICODE)'0') + ch/100;
							ch = ch % 100;
						}
						if (ch>10)
						{
							*p++ = ((Py_UNICODE)'0') + ch/10;
							ch = ch % 10;
						}
						*p++ = ((Py_UNICODE)'0') + ch;
						*p++ = (Py_UNICODE)';';
					}
					else
					{
						Py_DECREF(result);
						return NULL;
					}
				}
				else /* we were able to encode this character, so use it */
				{
					*p++ = ch;
					Py_DECREF(test);
				}
			}
			return (PyUnicodeObject *)result;
		}
		else /* a "real exception" */
			return NULL;
	}
	else /* no exception, i.e. it worked, so we can use the string as it is */
	{
		Py_DECREF(test); /* we no longer need the encoded version */
		/* we don't incref here, the caller knows this */
		return str;
	}
}

static PyObject *__escapeText2(PyUnicodeObject *unicode, const char *encoding, struct Replacement *replacements)
{
	int i;
	int oldsize = PyUnicode_GET_SIZE(unicode);
	int newsize = 0;

	int *toBeReplaced = NULL;
	int replaceCount;

	PyUnicodeObject *res = NULL;

	toBeReplaced = PyMem_New(int, oldsize);

	if (!toBeReplaced)
		goto error;

	replaceCount = 0;
	for (i = 0; i < oldsize; ++i)
	{
		PyObject *test;
		Py_UNICODE ch = PyUnicode_AS_UNICODE(unicode)[i];
		struct Replacement *rep;
		for (rep = replacements; rep->character; ++rep)
		{
			if (ch==rep->character)
			{
				newsize += 1+rep->size+1;
				goto testok;
			}
		}
		/* none of the special characters => try to encode it */
		test = PyUnicode_Encode(&ch, 1, encoding, NULL);

		if (test == NULL) /* exception */
		{
			if (PyErr_ExceptionMatches(PyExc_UnicodeError)) /* we must create a character reference for it */
			{
				PyErr_Clear();

				if (ch < 10)
					newsize += 2+1+1;
				else if (ch < 100)
					newsize += 2+2+1;
				else if (ch < 1000)
					newsize += 2+3+1;
				else if (ch < 10000)
					newsize += 2+4+1;
				else
					newsize += 2+5+1;
				toBeReplaced[replaceCount++] = i; /* remember position */
			}
			else /* unexpected encoding exception */
				goto error;
		}
		else /* we were able to encode this character, so use it */
		{
			++newsize;
			Py_DECREF(test);
		}

		testok:
		;
	}
	if (oldsize==newsize) /* nothing to replace? */
	{
		/* return original */
		Py_INCREF(unicode);
		res = unicode;
	}
	else
	{
		int *currentReplacer = toBeReplaced;
		Py_UNICODE *output;
		res = (PyUnicodeObject *)PyUnicode_FromUnicode(NULL, newsize);
		if (!res)
			goto error;
		output = PyUnicode_AS_UNICODE(res);

		for (i = 0; i < oldsize; ++i)
		{
			Py_UNICODE ch = PyUnicode_AS_UNICODE(unicode)[i];

			if (i==*currentReplacer)
			{
				*output++ = '&';
				*output++ = '#';
				if (ch>10000)
				{
					*output++ = ((Py_UNICODE)'0') + ch/10000;
					ch = ch % 10000;
				}
				if (ch>1000)
				{
					*output++ = ((Py_UNICODE)'0') + ch/1000;
					ch = ch % 1000;
				}
				if (ch>100)
				{
					*output++ = ((Py_UNICODE)'0') + ch/100;
					ch = ch % 100;
				}
				if (ch>10)
				{
					*output++ = ((Py_UNICODE)'0') + ch/10;
					ch = ch % 10;
				}
				*output++ = ((Py_UNICODE)'0') + ch;
				*output++ = ';';

				++currentReplacer;
			}
			else
			{
				struct Replacement *rep;
				for (rep = replacements; rep->character; ++rep)
				{
					if (ch==rep->character)
					{
						*output++ = '&';
						memcpy(output, rep->replacement, sizeof(Py_UNICODE)*rep->size);
						output += rep->size;
						*output++ = ';';

						goto encodeok;
					}
				}
				*output++ = ch;
			}
			encodeok:
			;
		}
	}
	goto ok;
	error:
	Py_XDECREF(res);
	ok:
	PyMem_Del(toBeReplaced);
	return (PyObject *)res;
}

static char escapeText__doc__[] =
"escapeText(unicode, encoding) -> unicode\n\
\n\
Return a copy of the string S, where every occurrence of\n\
'<', '>' and '&' and all unencodable characters in the\n\
specified encoding have been replaced with their XML character entity.";

static PyObject *escapeText(PyObject *self, PyObject *args)
{
	int i;
	int oldsize;
	int newsize = 0;
	PyUnicodeObject *str;
	const char *encoding = NULL;

	if (!PyArg_ParseTuple(args, "Os:escapeText", &str, &encoding))
		return NULL;
	str = (PyUnicodeObject *)PyUnicode_FromObject((PyObject *)str);
	if (str == NULL)
		return NULL;

	oldsize = PyUnicode_GET_SIZE(str);
	for (i = 0; i < oldsize; ++i)
	{
		Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
		if (ch == ((Py_UNICODE)'<'))
			newsize += COUNTOF(lt);
		else if (ch == ((Py_UNICODE)'>')) /* Note that we always replace '>' with its entity, not just in case it is part of ']]>' */
			newsize += COUNTOF(gt);
		else if (ch == ((Py_UNICODE)'&'))
			newsize += COUNTOF(amp);
		else
			newsize++;
	}
	if (oldsize==newsize)
	{
		/* nothing to replace, continue with charrefs */
		Py_INCREF(str);
		return (PyObject *)escapeUnencodable(str, encoding);
	}
	else
	{
		PyUnicodeObject *result = (PyUnicodeObject *)PyUnicode_FromUnicode(NULL, newsize);
		Py_UNICODE *p;
		if (result == NULL)
			return NULL;
		p = PyUnicode_AS_UNICODE(result);
		for (i = 0; i < oldsize; ++i)
		{
			Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
			if (ch == ((Py_UNICODE)'<'))
			{
				Py_UNICODE_COPY(p, lt, COUNTOF(lt));
				p += COUNTOF(lt);
			}
			else if (ch == ((Py_UNICODE)'>'))
			{
				Py_UNICODE_COPY(p, gt, COUNTOF(gt));
				p += COUNTOF(gt);
			}
			else if (ch == ((Py_UNICODE)'&'))
			{
				Py_UNICODE_COPY(p, amp, COUNTOF(amp));
				p += COUNTOF(amp);
			}
			else
				*p++ = ch;
		}
		return (PyObject *)escapeUnencodable(result, encoding);
	}
}

static PyObject *escapeText2(PyObject *self, PyObject *args)
{
	PyUnicodeObject *unicode;
	const char *encoding = NULL;

	if (!PyArg_ParseTuple(args, "O!s:escapeText", &unicode, &encoding))
		return NULL;

	return __escapeText2(unicode, encoding, &replacements[1]);
}

static char escapeAttr__doc__[] =
"escapeAttr(unicode, encoding) -> unicode\n\
\n\
Return a copy of the string S, where every occurrence of\n\
'<', '>', '&', and '\"' and all unencodable characters in the\n\
specified encoding have been replaced with their XML character entity.";

static PyObject *escapeAttr(PyObject *self, PyObject *args)
{
	int i;
	int oldsize;
	int newsize = 0;
	PyUnicodeObject *str;
	const char *encoding;

	if (!PyArg_ParseTuple(args, "Os:escapeAttr", &str, &encoding))
		return NULL;
	str = (PyUnicodeObject *)PyUnicode_FromObject((PyObject *)str);
	if (str == NULL)
		return NULL;

	oldsize = PyUnicode_GET_SIZE(str);
	for (i = 0; i < oldsize; ++i)
	{
		Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
		if (ch == ((Py_UNICODE)'<'))
			newsize += COUNTOF(lt);
		else if (ch == ((Py_UNICODE)'>'))
			newsize += COUNTOF(gt);
		else if (ch == ((Py_UNICODE)'&'))
			newsize += COUNTOF(amp);
		else if (ch == ((Py_UNICODE)'\"'))
			newsize += COUNTOF(quot);
		else
			newsize++;
	}
	if (oldsize==newsize)
	{
		/* nothing to replace, continue with charrefs */
		Py_INCREF(str);
		return (PyObject *)escapeUnencodable(str, encoding);
	}
	else
	{
		PyUnicodeObject *result = (PyUnicodeObject *)PyUnicode_FromUnicode(NULL, newsize);
		Py_UNICODE *p;
		if (result == NULL)
			return NULL;
		p = PyUnicode_AS_UNICODE(result);
		for (i = 0; i < oldsize; ++i)
		{
			Py_UNICODE ch = PyUnicode_AS_UNICODE(str)[i];
			if (ch == ((Py_UNICODE)'<'))
			{
				Py_UNICODE_COPY(p, lt, COUNTOF(lt));
				p += COUNTOF(lt);
			}
			else if (ch == ((Py_UNICODE)'>'))
			{
				Py_UNICODE_COPY(p, gt, COUNTOF(gt));
				p += COUNTOF(gt);
			}
			else if (ch == ((Py_UNICODE)'&'))
			{
				Py_UNICODE_COPY(p, amp, COUNTOF(amp));
				p += COUNTOF(amp);
			}
			else if (ch == ((Py_UNICODE)'"'))
			{
				Py_UNICODE_COPY(p, quot, COUNTOF(quot));
				p += COUNTOF(quot);
			}
			else
				*p++ = ch;
		}
		return (PyObject *)escapeUnencodable(result, encoding);
	}
}

/* reimplement unistr, which has been removed before the release of Python 2.1 */
static PyObject *builtin_unistr(PyObject *self, PyObject *args)
{
	PyObject *v;

	if (!PyArg_ParseTuple(args, "O:unistr", &v))
		return NULL;
	return PyObject_Unicode(v);
}

static char unistr_doc[] =
"unistr(object) -> unicode\n\
\n\
Return a nice unicode representation of the object.\n\
If the argument is a unicode, the return value is the same object.";

/* ==================================================================== */
/* python module interface */

static PyMethodDef _functions[] =
{
	{"escapeText2", escapeText,     METH_VARARGS, escapeText__doc__},
	{"escapeText",  escapeText,     METH_VARARGS, escapeText__doc__},
	{"escapeAttr",  escapeAttr,     METH_VARARGS, escapeAttr__doc__},
	{"unistr",      builtin_unistr, METH_VARARGS, unistr_doc},
	{NULL, NULL}
};

void
#ifdef WIN32
__declspec(dllexport)
#endif
inithelpers(void)
{
	Py_InitModule("helpers", _functions);
}
